-- stablesort (c) 2013 Dirk Laurie and Steve Fisher
-- Posted on Lua-L by Steve Fisher on 02 May 2013 22:10:50 
-- Modularized an critical functions implemented in C API by DPL

-- Optimal max_chunk_size depends on many things: typical data, properties
--   of your system etc. For Lus 5.2 on a typical laptop, Steve recommends:
local max_chunk_size = 12
-- If you're using LuaJIT, Steve recommends 72, but that was for the pure
-- Lua version. This version has not been ported to LuaJIT yet. 

-- if a global or upvalue core is given, it is not reloaded

local core=core or require"sort_core"
local small_sort = core.sort
local merge = core.merge

local function merge_sort( array, low, high, goes_before )
  if high - low < max_chunk_size then
    small_sort( array, low, high, goes_before )
  else
    local middle = math.floor((low + high)/2)
    merge_sort( array, low, middle, goes_before )
    merge_sort( array, middle + 1, high, goes_before )
    merge( array, low, middle, high, goes_before )
  end
end

local function stablesort( array, goes_before )
  local n = #array
  if n < 2 then  return array  end
  if goes_before and goes_before(array[1],array[1]) then 
    error"invalid order function for sorting"
  end
  merge_sort( array, 1, n, goes_before )
  return array
end

return stablesort
